function random_walk_metropolis_hastings(TargetFun,ProposalFun,xparam1,vv,mh_bounds,varargin)
%function random_walk_metropolis_hastings(TargetFun,ProposalFun,xparam1,vv,mh_bounds,varargin)
% Random walk Metropolis-Hastings algorithm. 
% 
% INPUTS 
%   o TargetFun  [char]     string specifying the name of the objective
%                           function (posterior kernel).
%   o xparam1    [double]   (p*1) vector of parameters to be estimated (initial values).
%   o vv         [double]   (p*p) matrix, posterior covariance matrix (at the mode).
%   o mh_bounds  [double]   (p*2) matrix defining lower and upper bounds for the parameters. 
%   o varargin              list of argument following mh_bounds
%  
% OUTPUTS 
%   None  
%
% ALGORITHM 
%   Metropolis-Hastings.       
%
% SPECIAL REQUIREMENTS
%   None.

% Copyright (C) 2006-2008 Dynare Team
%
% This file is part of Dynare.
%
% Dynare is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% Dynare is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with Dynare.  If not, see <http://www.gnu.org/licenses/>.

global M_ options_ bayestopt_	estim_params_	trend_coeff_	oo_  ...
		xparam1_test it_;

%%%%
%%%% Initialization of the random walk metropolis-hastings chains.
%%%%
[ ix2, ilogpo2, ModelName, MhDirectoryName, fblck, fline, npar, nblck, nruns, NewFile, MAX_nruns, d ] = ...
    metropolis_hastings_initialization(TargetFun, xparam1, vv, mh_bounds, varargin{:});
options_.lik_algo = 1;
OpenOldFile = ones(nblck,1);
if strcmpi(ProposalFun,'rand_multivariate_normal')
    n = npar;
elseif strcmpi(ProposalFun,'rand_multivariate_student')
    n = options_.student_degrees_of_freedom;
end
load([MhDirectoryName '/' ModelName '_mh_history.mat'],'record');
%%%%
%%%% NOW i run the (nblck-fblck+1) metropolis-hastings chains
%%%%
InitSizeArray = min([repmat(MAX_nruns,nblck,1) fline+nruns-1],[],2);
jscale = diag(bayestopt_.jscale);




if(fblck > 1)
	error('Need to code crash resumption properly for the parallel implementation. Sorry. --JH');
end


if(matlabpool('size')==0) 		% If there are no labs running...
    matlabpool open;				% then start some
end



if(nblck > 1)

	% Several chains are being run in parallel. Use multicore to split the workload.
	
	parmlist = cell(1,nblck);
	if(strcmp(TargetFun, 'DsgeVarLikelihood'))
		%DSGE-BVAR stuff
		mYY = evalin('base', 'mYY');
		mYX = evalin('base', 'mYX');
		mXY = evalin('base', 'mXY');
		mXX = evalin('base', 'mXX');

		for k = 1:nblck
			parmlist{k} = {k, ix2, ilogpo2, ModelName, MhDirectoryName, fblck, fline, npar, nblck, nruns, NewFile, MAX_nruns, d,  ...
							 ProposalFun, TargetFun, mh_bounds, jscale, n, amh_t0, amh_eps, d_saved, ...
							 bayestopt_, estim_params_, options_, trend_coeff_, M_, oo_, xparam1_test, it_, ...
							 mYY, mYX, mXY, mXX, varargin{:}}; 
		end
	else
		for k = 1:nblck
			parmlist{k} = {k, ix2, ilogpo2, ModelName, MhDirectoryName, fblck, fline, npar, nblck, nruns, NewFile, MAX_nruns, d,  ...
							 ProposalFun, TargetFun, mh_bounds, jscale, n, amh_t0, amh_eps, d_saved, ...
							 bayestopt_, estim_params_, options_, trend_coeff_, M_, oo_, xparam1_test, it_, ...
							 varargin{:}}; 
		end
	end


	% MaxMasterEvaluations (the fourth argument to startmulticoremaster) is
	% set to one, because otherwise the RWMH chains often finish at around
	% the same time; if the master finishes slightly before the slave, the
	% master will not see a result file and not realise that the slave has
	% already done the second chain...
	startmulticoremaster(@inner_RWMH, parmlist, 'multicoref', 1);
else

	% It's just one chain. We'll use the parallel toolbox alone.
	
	if(strcmp(TargetFun, 'DsgeVarLikelihood'))
		%DSGE-BVAR stuff
		mYY = evalin('base', 'mYY');
		mYX = evalin('base', 'mYX');
		mXY = evalin('base', 'mXY');
		mXX = evalin('base', 'mXX');
		inner_RWMH(1, ix2, ilogpo2, ModelName, MhDirectoryName, fblck, fline, npar, nblck, ...
			nruns, NewFile, MAX_nruns, d, ProposalFun, TargetFun, mh_bounds, jscale, ...
			n, amh_t0, amh_eps, d_saved, bayestopt_, estim_params_, options_, ...
			trend_coeff_, M_, oo_,  xparam1_test, it_, ...
			mYY, mYX, mXY, mXX, varargin{:});
	else
		inner_RWMH(1, ix2, ilogpo2, ModelName, MhDirectoryName, fblck, fline, npar, nblck, ...
			nruns, NewFile, MAX_nruns, d, ProposalFun, TargetFun, mh_bounds, jscale, ...
			n, amh_t0, amh_eps, d_saved, bayestopt_, estim_params_, options_, ...
			trend_coeff_, M_, oo_, xparam1_test, it_, ...
				varargin{:});
	end
end


% record.Seeds.Normal = randn('state');
% record.Seeds.Unifor = rand('state');
save([MhDirectoryName '/' ModelName '_mh_history.mat'],'record');
disp(['MH: Number of mh files			: ' int2str(NewFile(1)) ' per block.'])
disp(['MH: Total number of generated files	: ' int2str(NewFile(1)*nblck) '.'])
% disp(['MH: Total number of iterations 		: ' int2str((NewFile(1)-1)*MAX_nruns+irun-1) '.'])
disp('MH: average acceptation rate per chain   : ')
% disp(record.AcceptationRates);
disp(' ')